/* Mptn -- regexp-like pattern matcher suited for morphological analysis
 * Copyright (C) 2000  George Bronnikov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
  A set of bindings for Mptn library for MzScheme.
*/

#include <escheme.h>
#include "mptn.h"

/* Global vars */
static mptn_control_t *mz_mptn_control;
static Scheme_Type mz_mptn_type;
static Scheme_Type mz_ectl_type;
static Scheme_Type mz_matcher_type;

static Scheme_Object *sym_valid;
static Scheme_Object *sym_vars;

static Scheme_Object *sym_init;
static Scheme_Object *sym_destroy;
static Scheme_Object *sym_iter_start;
static Scheme_Object *sym_iter_step;
static Scheme_Object *sym_ictl_free;
static Scheme_Object *sym_stage;
static Scheme_Object *sym_subst;
static Scheme_Object *sym_print;

/**********************************  Type mptn ********************************/

typedef struct mz_mptn mz_mptn_t;
struct mz_mptn {
  Scheme_Type type;
  mptn_t *mptn;
};

#define SCHEME_MPTN_VAL(x) (((mz_mptn_t*)(x))->mptn)
#define SCHEME_MPTNP(x) (SCHEME_TYPE(x) == mz_mptn_type)

static void mz_mptn_finalizer (void *obj, void *_);

/* Producing an Mptn object by parsing a string 
   (mptn-parse expr)
*/
Scheme_Object *
mz_mptn_parse (int argc, Scheme_Object *argv[])
{
  mz_mptn_t *res;
  mptn_t *res_mptn;

  g_assert (argc == 1);

  if (!SCHEME_STRINGP (argv[0]))
    scheme_signal_error ("mptn-parse: need a string");

  res_mptn = mptn_parse (mz_mptn_control, SCHEME_STR_VAL (argv[0]));
  if (!res_mptn)
    scheme_signal_error ("mptn-parse: parse error");

  /* Allocate the resulting object */
  res = scheme_malloc (sizeof (mz_mptn_t));
  res->type = mz_mptn_type;
  res->mptn = res_mptn;
  
  /* Make sure it gets properly deallocated */
  scheme_register_finalizer (res, mz_mptn_finalizer, NULL, NULL, NULL);

  return (Scheme_Object *)res;
}

static void
mz_mptn_finalizer (void *p, void *_)
{
  mz_mptn_t *mz_mptn = p;
  mptn_free (mz_mptn->mptn);
}

/*
  Check for the type mptn
  (mptn? x)
*/
static Scheme_Object *
mz_mptn_p (int argc, Scheme_Object *argv[])
{
  g_assert (argc == 1);

  if (SCHEME_MPTNP (argv[0]))
    return scheme_true;
  else
    return scheme_false;
}


/*********************** Mptn ectl / variable bindings ************************/

/* The lifetime of the object referenced by this type depends on which mptn calls
   gett issued.  It may become invalid.
   
   This same type is used to represent variables passed to mptn_matcher_iter_start_ft
*/

typedef struct mz_ectl mz_ectl_t;
struct mz_ectl {
  Scheme_Type type;
  mptn_ectl_t *ectl;
  mptn_var_t *vars;
  char *str;
};

#define SCHEME_MPTN_ECTLP(x) (SCHEME_TYPE(x) == mz_ectl_type)
#define SCHEME_MPTN_ECTL_VAL(x) (((mz_ectl_t*)(x))->ectl)
#define SCHEME_MPTN_ECTL_VARS(x) (((mz_ectl_t*)(x))->vars)

#define SCHEME_MPTN_ECTL_SET(x,e) (((mz_ectl_t*)(x))->ectl = (e),			\
				   ((mz_ectl_t*)(x))->vars = (e) ? mptn_ectl_vars (e) : NULL)

#define SCHEME_MPTN_ECTL_SET_VARS(x,v) (((mz_ectl_t*)(x))->ectl = NULL,			\
					((mz_ectl_t*)(x))->vars = v)

static void mz_ectl_finalizer (void *p, void *_);

/* 
   Internal function to wrap mptn_ectl into a Scheme object.
   (also saves the string to prevent it from gc).
   Returns #f if ectl is NULL.
*/
static Scheme_Object *
mz_make_ectl (mptn_ectl_t *ectl, char *str)
{
  Scheme_Object *rr;
  mz_ectl_t *res;

  if (ectl)
    {
      res = scheme_malloc (sizeof(mz_ectl_t));
      res->type = mz_ectl_type;
      res->ectl = ectl;
      res->vars = mptn_ectl_vars (ectl);
      res->str = str;

      /* Make sure it gets properly deallocated */
      scheme_register_finalizer (res, mz_ectl_finalizer, NULL, NULL, NULL);

      rr = (Scheme_Object*) res;
    }
  else
    rr = scheme_false;

  return rr;
}

/* Another way to create an mz_ectl_t: no ectl proper, just variables. */
static Scheme_Object *
mz_make_vars_ectl (mptn_var_t *vars, char *str)
{
  mz_ectl_t *res;

  res = scheme_malloc (sizeof(mz_ectl_t));
  res->type = mz_ectl_type;
  res->ectl = NULL;
  res->vars = vars;
  res->str = str;

  /* Make sure it gets properly deallocated */
  scheme_register_finalizer (res, mz_ectl_finalizer, NULL, NULL, NULL);

  return (Scheme_Object*)res;
}

static void 
mz_ectl_finalizer (void *p, void *_)
{
  mz_ectl_t *arg = p;

  g_assert (arg->type == mz_ectl_type);

  if (arg->ectl)
    mptn_ectl_free (arg->ectl);
}

/*
  Type testing predicate.
  (mptn-ectl? x)
*/
static Scheme_Object *
mz_mptn_ectl_p (int argc, Scheme_Object *argv[])
{
  g_assert (argc == 1);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    return scheme_true;
  else
    return scheme_false;
}

/*
  Test is an ectl is still valid.
  (mptn-ectl-valid? ectl)
  
  Returns:
  	valid -- ectl is valid
	vars  -- ectl is for variable access only
	#f    -- ectl is not valid
*/
static Scheme_Object *
mz_mptn_ectl_valid_p (int argc, Scheme_Object *argv[])
{
  mptn_ectl_t *ectl;
  mptn_var_t *vars;
  
  g_assert (argc = 1);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    {
      ectl = SCHEME_MPTN_ECTL_VAL (argv[0]);
      vars = SCHEME_MPTN_ECTL_VARS (argv[0]);
    }
  else
    scheme_signal_error ("mptn-ectl-valid?: expect ectl as argument");

  if (ectl)
    return sym_valid;
  else if (vars)
    return sym_vars;
  else
    return scheme_false;
}

/* Get a value of a variable -- variable name may given as symbol or string
   (mptn-ectl-ref ectl vname)
*/
static Scheme_Object *
mz_mptn_ectl_ref (int argc, Scheme_Object *argv[])
{
  mptn_var_t *vars;
  GQuark quark;

  g_assert (argc == 2);
  
  if (!SCHEME_MPTN_ECTLP (argv[0]))
    scheme_signal_error ("mptn-ectl-ref: expected mptn-ectl as the 1st argument");
  vars = SCHEME_MPTN_ECTL_VARS(argv[0]);

  if (!vars)
    scheme_signal_error ("mptn-ectl-ref: ectl invalid");

  if (SCHEME_STRINGP (argv[1]))
    quark = g_quark_from_string (SCHEME_STR_VAL(argv[1]));
  else if (SCHEME_SYMBOLP (argv[1]))
    quark = g_quark_from_string (SCHEME_SYM_VAL (argv[1])); 	/* case? */
  else
    scheme_signal_error ("mptn-ectl-ref: expected string or symbol as the 2nd argument");

  for (; vars->vcode; vars++)
    if (vars->vcode == quark && vars->beg)		/* found; the var is set */
      return scheme_make_sized_string (vars->beg, vars->end - vars->beg, TRUE);
    else if (vars->vcode == quark && !vars->beg) 	/* found; the var is not set */
      return scheme_false;

  return scheme_false;		/* variable not found */
}

/*  Convert mptn-ectl into an alist.
    (mptn-ectl->alist ectl)
*/
static Scheme_Object *
mz_mptn_ectl_to_alist (int argc, Scheme_Object *argv[])
{
  Scheme_Object *res;
  mptn_var_t *vars;

  g_assert (argc == 1);

  if (!SCHEME_MPTN_ECTLP (argv[0]))
    scheme_signal_error ("mptn-ectl->alist: expect an mptn-ectl argument");
  vars = SCHEME_MPTN_ECTL_VARS (argv[0]);

  if (!vars)
    scheme_signal_error ("mptn-ectl-ref: ectl invalid");

  res = scheme_null;
  for (; vars->vcode; vars++)
    if (vars->beg)
      res = scheme_make_pair (scheme_make_pair (scheme_intern_symbol (g_quark_to_string (vars->vcode)),
						scheme_make_sized_string (vars->beg, vars->end - vars->beg, TRUE)),
			      res);

  return res;
}

/* Does Ectl have a place for Var?
   (mptn-ectl-has-var? ectl var)
*/
static Scheme_Object *
mz_mptn_ectl_has_var_p (int argc, Scheme_Object* argv[])
{
  mptn_var_t *vars;
  GQuark quark;

  g_assert (argc == 2);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    vars = SCHEME_MPTN_ECTL_VARS(argv[0]);
  else
    scheme_signal_error ("mptn-ectl-has-var?: expect mptn-ectl as the 1st argument");

  if (!vars)
    scheme_signal_error ("mptn-ectl-has-var?: invalid mptn-ectl");

  if (SCHEME_SYMBOLP (argv[1]))
    quark = g_quark_from_string (SCHEME_SYM_VAL (argv[1]));
  else if (SCHEME_STRINGP (argv[1]))
    quark = g_quark_from_string (SCHEME_STR_VAL (argv[1]));
  else
    scheme_signal_error ("mptn-ectl-has-var?: expect symbol or string as the 2nd argument");

  for (; vars->vcode; vars++)
    if (vars->vcode == quark)
      return scheme_true;

  return scheme_false;
}

/* Set a variable value inside Ectl.
   (mptn-ectl-set! ectl var str)
*/
static Scheme_Object*
mz_mptn_ectl_set_bang (int argc, Scheme_Object *argv[])
{
  mptn_ectl_t *ectl;
  mptn_var_t *vars;
  char *str;
  GQuark quark;

  g_assert (argc == 3);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    {
      ectl = SCHEME_MPTN_ECTL_VAL (argv[0]);
      vars = SCHEME_MPTN_ECTL_VARS(argv[0]);
    }
  else
    scheme_signal_error ("mptn-ectl-set!: expect mptn-ectl as the 1st argument");

  if (!vars)
    scheme_signal_error ("mptn-ectl-set!: invalid mptn-ectl");

  if (SCHEME_SYMBOLP (argv[1]))
    quark = g_quark_from_string (SCHEME_SYM_VAL (argv[1]));
  else if (SCHEME_STRINGP (argv[1]))
    quark = g_quark_from_string (SCHEME_STR_VAL (argv[1]));
  else
    scheme_signal_error ("mptn-ectl-set!: expect symbol or string as the 2nd argument");

  if (SCHEME_STRINGP (argv[2]))
    str = SCHEME_STR_VAL (argv[2]);
  else
    scheme_signal_error ("mptn-ectl-set!: expect string as the 3rd argument");

  for (; vars->vcode; vars++)
    if (vars->vcode == quark)
      {
	/* Found; set the new value */

	if (vars->beg && vars->allocated)	/* May need to free the old value */
	  g_free (vars->beg);

	vars->beg = chra_dup (str);
	vars->end = vars->beg + chra_len (vars->beg);
	vars->allocated = TRUE;
	vars->stage = ectl ? mptn_ectl_stage (ectl) : 0;
	return scheme_undefined;
      }

  scheme_signal_error ("mptn-ectl-set!: variable not present in ectl");
}

/* Get the current stage of an ectl.
   (mptn-ectl->stage ectl)
*/
static Scheme_Object *
mz_mptn_ectl_stage (int argc, Scheme_Object *argv[])
{
  mptn_ectl_t *ectl;
  
  g_assert (argc == 1);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    ectl = SCHEME_MPTN_ECTL_VAL (argv[0]);
  else
    scheme_signal_error ("mptn-ectl->stage: expect mptn-ectl as argument");

  if (!ectl)
    scheme_signal_error ("mptn-ectl-ref: ectl invalid");

  return scheme_make_integer (mptn_ectl_stage (ectl));
}

/*
  Step an ectl.
  (mptn-ectl-step! ectl)
  
  Moves ectl, modifies the mz_ectl object given as parameter and retrns it.
  If the loop ends, returns #f.
*/
static Scheme_Object *
mz_mptn_ectl_step_bang (int argc, Scheme_Object *argv[])
{
  mptn_ectl_t *ectl;
  
  g_assert (argc == 1);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    ectl = SCHEME_MPTN_ECTL_VAL (argv[0]);
  else
    scheme_signal_error ("mptn-ectl-step!: expect mptn-ectl as argument");

  if (!ectl)
    scheme_signal_error ("mptn-ectl-ref: ectl invalid");

  ectl = mptn_exec_step (ectl);

  /* Modify the argument */
  SCHEME_MPTN_ECTL_SET (argv[0], ectl);

  if (!ectl)
    ((mz_ectl_t*)argv[0])->str = NULL;	/* No more need to gc-protect */

  if (ectl)
    return argv[0];
  else
    return scheme_false;
}

/*
  Destroy the ectl.
  (mptn-ectl-destroy! ectl)
*/
static Scheme_Object *
mz_mptn_ectl_destroy_bang (int argc, Scheme_Object *argv[])
{
  mptn_ectl_t *ectl;
  
  g_assert (argc == 1);

  if (SCHEME_MPTN_ECTLP (argv[0]))
    ectl = SCHEME_MPTN_ECTL_VAL (argv[0]);
  else
    scheme_signal_error ("mptn-ectl-destroy!: expect mptn-ectl as argument");

  if (ectl)
    mptn_ectl_free (ectl);

  /* Modify the argument */
  SCHEME_MPTN_ECTL_SET (argv[0], NULL);
  ((mz_ectl_t*)argv[0])->str = NULL;  /* No longer needed */

  return scheme_void;
}


/* Pattern matching procedure 
   (mptn-exec-for-each mptn str data proc)
   The procedure should take 1 argument -- vars assignment.
   !!!! The assignment cannot be stored in any long-lived location 
   (say, a variable with a bigger scope than Proc).
*/
static Scheme_Object *
mz_mptn_exec_loop (int argc, Scheme_Object *argv[], gboolean map)
{
  mptn_t *mptn;
  char *str;
  Scheme_Object *data;
  Scheme_Object *proc;
  Scheme_Object *tt;

  mptn_ectl_t *ectl;
  Scheme_Object *res;
  Scheme_Object **rp;
  
  g_assert (argc == 4);

  if (!SCHEME_MPTNP(argv[0]))
    scheme_signal_error ("mptn-exec-loop: expect mptn as the 1st argument");
  mptn = SCHEME_MPTN_VAL (argv[0]);

  if (!SCHEME_STRINGP (argv[1]))
    scheme_signal_error ("mptn-exec-loop: expect string as the 2nd argument");
  str = SCHEME_STR_VAL (argv[1]);

  data = argv[2];

  if (!SCHEME_PROCP (argv[3]))
    scheme_signal_error ("mptn-exec-loop: expect procedure as the 4th argument");
  proc = argv[3];

  /* prepare for the result construction */
  res = map ? scheme_null : scheme_false;
  rp = &res;

  /* !! We need to work with just one Scheme ectl object.
     Otherwise the system will attempt to gc the others sooner or later
     and free invalid pointers.
  */
  for (ectl = mptn_exec_start (mptn, str, NULL, data),
	 tt = mz_make_ectl (ectl, str);
       ectl;
       ectl = mptn_exec_step (ectl),
	 SCHEME_MPTN_ECTL_SET (tt, ectl))
    {
      Scheme_Object *tr;
      tr = _scheme_apply (proc, 1, &tt);

      /* If we need a map of all results, then make another element of it */
      if (map)
	{
	  *rp = scheme_make_pair (tr, scheme_null);
	  rp = &SCHEME_CDR (*rp);
	}
      else
	res = scheme_true;
    }

  return res;
}

static Scheme_Object *
mz_mptn_exec_for_each (int argc, Scheme_Object *argv[])
{
  return mz_mptn_exec_loop (argc, argv, FALSE);
}

static Scheme_Object *
mz_mptn_exec_map (int argc, Scheme_Object *argv[])
{
  return mz_mptn_exec_loop (argc, argv, TRUE);
}

/*
  "Lower level" pattern matching procedure -- just a translation from C.
  (mptn-exec-start mptn str [data])
  Returns ectl.
*/
static Scheme_Object *
mz_mptn_exec_start (int argc, Scheme_Object *argv[])
{
  mptn_t *mptn;
  char *str;
  Scheme_Object *data;

  mptn_ectl_t *ectl;
  Scheme_Object *res;
  
  g_assert (2 <= argc && argc <= 3);

  if (!SCHEME_MPTNP(argv[0]))
    scheme_signal_error ("mptn-exec-start: expect mptn as the 1st argument");
  mptn = SCHEME_MPTN_VAL (argv[0]);

  if (!SCHEME_STRINGP (argv[1]))
    scheme_signal_error ("mptn-exec-start: expect string as the 2nd argument");
  str = SCHEME_STR_VAL (argv[1]);

  data = (argc == 3) ? argv[2] : scheme_false;

  ectl = mptn_exec_start (mptn, str, NULL, data);
  if (ectl)
    res = mz_make_ectl (ectl, str);
  else
    res = scheme_false;

  return res;
}

/* "Internal" variant of the same procedure
   (mptn-exec-start-with-ectl-and-stage mptn str ectl stage [data]).
 */
static Scheme_Object *
mz_mptn_exec_start_with_ectl_and_stage (int argc, Scheme_Object *argv[])
{
  mptn_t *mptn;
  char *str;
  mptn_var_t *vars;
  int stage;
  Scheme_Object *data;

  mptn_ectl_t *ectl;
  Scheme_Object *res;

  g_assert (4 <= argc && argc <= 5);

  if (!SCHEME_MPTNP(argv[0]))
    scheme_signal_error ("mptn-exec-start-with-ectl-and-stage: expected mptn as the 1st argument");
  mptn = SCHEME_MPTN_VAL (argv[0]);

  if (!SCHEME_STRINGP(argv[1]))
    scheme_signal_error ("mptn-exec-start-with-ectl-and-stage: expected string as the 2nd argument");
  str = SCHEME_STR_VAL (argv[1]);

  if (!SCHEME_MPTN_ECTLP(argv[2]))
    scheme_signal_error ("mptn-exec-start-with-ectl-and-stage: expected mptn-ectl as the 3rd argument");
  vars = SCHEME_MPTN_ECTL_VARS (argv[2]);

  if (!SCHEME_INTP(argv[3]))
    scheme_signal_error ("mptn-exec-start-with-ectl-and-stage: expected integer as the 4th argument");
  stage = SCHEME_INT_VAL (argv[3]);

  data = (argc == 5) ? argv[4] : scheme_false;
  

  ectl = mptn_exec_start_with_vars (mptn, 
				    str, NULL, 
				    vars, stage,
				    data);
  if (ectl)
    res = mz_make_ectl (ectl, str);
  else
    res = scheme_false;

  return res;
}

/* Substituting variable values into Mptn.
   (mptn-subst mptn ectl-vars data)
   Returns a string or #f on failure.
*/
static Scheme_Object *
mz_mptn_subst (int argc, Scheme_Object *argv[])
{
  mptn_t *mptn;
  mptn_var_t *vars;
  Scheme_Object *data;

  char *strres;
  Scheme_Object *res;
  
  g_assert (argc >= 2 && argc <= 3);

  if (!SCHEME_MPTNP(argv[0]))
    scheme_signal_error ("mptn-subst: expect mptn as the 1st argument");
  mptn = SCHEME_MPTN_VAL (argv[0]);

  if (!SCHEME_MPTN_ECTLP (argv[1]))
    scheme_signal_error ("mptn-subst: expect mptn-ectl as the 2nd argument");
  vars = SCHEME_MPTN_ECTL_VARS (argv[1]);

  data = (argc == 3) ? argv[2] : scheme_false;

  strres = mptn_subst (mptn, vars, data);

  if (!strres)
    return scheme_false;
  else
    {
      res = scheme_make_string (strres);
      g_free (strres);	/* was allocated inside mptn_subst */
      return res;
    }
}


/*************************** Writing matchers in Scheme ***********************/
typedef struct mz_matcher mz_matcher_t;

struct mz_matcher {
  Scheme_Type type;
  Scheme_Object *init;
  Scheme_Object *destroy;
  Scheme_Object *iter_start;
  Scheme_Object *iter_step;
  Scheme_Object *ictl_free;
  Scheme_Object *stage;
  Scheme_Object *subst;
  Scheme_Object *print;
};

/* Make a matcher.
   (mptn-matcher 'init init 'destoy destroy ...)
*/

static Scheme_Object *
mz_mptn_make_matcher (int argc, Scheme_Object *argv[])
{
  mz_matcher_t *mzm;

  /* Allocate the structure. Will be referenced from the mathcre object */
  mzm = scheme_malloc (sizeof(mz_matcher_t));
  memset (mzm, 0, sizeof (mz_matcher_t));
  mzm->type = mz_matcher_type;

  /* Fill in the structure */
  for(; argc > 0; argc -=2, argv += 2)
    {
      if (argc == 1 || !SCHEME_SYMBOLP(argv[0]))
	scheme_signal_error ("mptn-matcher: arguments should come in keyword-value pairs");

      if (argv[0] == sym_init)
	mzm->init = argv[1];
      else if (argv[0] == sym_destroy)
	mzm->destroy = argv [1];
      else if (argv[0] == sym_iter_start)
	mzm->iter_start = argv[1];
      else if (argv[0] == sym_iter_step)
	mzm->iter_step = argv[1];
      else if (argv[0] == sym_ictl_free)
	mzm->ictl_free = argv[1];
      else if (argv[0] == sym_stage)
	mzm->stage = argv[1];
      else if (argv[0] == sym_subst)
	mzm->subst = argv[1];
      else if (argv[0] == sym_print)
	mzm->print = argv[1];
      else
	{
	  char *fmt = "~s";
	  scheme_signal_error ("mptn-matcher: unknown keyword %s", 
			       scheme_format (fmt, strlen (fmt), 1, &argv[0], NULL));
	}
    }

  return (Scheme_Object*) mzm;
}


/***************** The matcher procedures interfacing to Scheme ****************/

typedef struct mz_ctl mz_ctl_t;
struct mz_ctl {
  mz_matcher_t *matcher;
  Scheme_Object *s_ctl;
};

/* Add a varname to varnames list */
static void
mz_add_var (GArray *varnames, Scheme_Object *v)
{
  char *str;
  GQuark quark;

  if (SCHEME_SYMBOLP (v))
    str = SCHEME_SYM_VAL (v);
  else if (SCHEME_STRINGP (v))
    str = SCHEME_STR_VAL (v);
  else
    scheme_signal_error ("mz-matcher matcher-init: variable name not symbol or string");

  quark = g_quark_from_string (str);

  {
    int i;
    gboolean found = FALSE;

    for (i = 0; i < varnames->len; i++)
      if (g_array_index (varnames, GQuark, i) == quark)
	{
	  found = TRUE;
	  break;
	}

    if (!found)
      g_array_append_val (varnames, quark);
  }
}

/* Init for Scheme matchers.
   Should have the form (init args udata).
   
   Should return (values mctl min-len max-len new-vars)
   min-len and max-len may be integers or #f.
   New-vars is a list of variable names used by the matcher.

   You may drop the later elements of the values list.
*/
static mptn_t *
mz_matcher_init (mptn_control_t *control,
		      chr *args_begin, chr *args_end,
		      GArray *varnames,
		      gpointer data)
{
  Scheme_Object *args;
  mz_ctl_t *mc = data;
  mz_matcher_t *mz_matcher = mc->matcher;
  Scheme_Object *udata = mc->s_ctl;

  /* The values for the mptn we will return */
  mptn_t *mptn;
  Scheme_Object * s_mctl = scheme_false;
  guint min_len = 0;
  guint max_len = G_MAXINT;

  /* Creating a string of arguments */
  if (!args_begin)
    args = scheme_false;
  else if (!args_end)
    args = scheme_make_string (args_begin);
  else
    args = scheme_make_sized_string (args_begin, args_end - args_begin, TRUE);

  if (mz_matcher->init)
    {
      Scheme_Object *mi_res;

      /* Now prepare the arguments and call the Scheme matcher-init function */
      Scheme_Object *mi_argv[2];
      
      mi_argv[0] = args;
      mi_argv[1] = udata;

      mi_res = _scheme_apply_multi (mz_matcher->init, 2, mi_argv);

      if (mi_res != scheme_multiple_values)
	s_mctl = mi_res;
      else
	{
	  if (scheme_multiple_count > 4)
	    scheme_signal_error ("mz-matcher matcher-init: should return (values mctl mun-len max-len vars), got %d values", scheme_multiple_count);
	  
	  s_mctl= scheme_multiple_array [0];

	  if (scheme_multiple_count > 1 && scheme_multiple_array[1] != scheme_false)
	    if (!SCHEME_INTP (scheme_multiple_array[1]))
	      scheme_signal_error ("mz-matcher matcher-init: min-len not integer");
	    else
	      min_len = SCHEME_INT_VAL (scheme_multiple_array[1]);

	  if (scheme_multiple_count > 2 && scheme_multiple_array[2] != scheme_false)
	    if (!SCHEME_INTP (scheme_multiple_array[2]))
	      scheme_signal_error ("mz-matcher matcher-init: max-len not integer");
	    else
	      max_len = SCHEME_INT_VAL (scheme_multiple_array[2]);

	  if (scheme_multiple_count > 3)
	    {
	      Scheme_Object *vlist;

	      /* The 4th member of the return values must be a list of varnames */
	      for (vlist = scheme_multiple_array[3];
		   vlist != scheme_null;
		   vlist = SCHEME_CDR (vlist))
		mz_add_var (varnames, SCHEME_CAR (vlist));
	    }
	}
    }

  /* Construct the Mptn */
  {
    struct mz_ctl *mc;

    mc = g_new0 (mz_ctl_t, 1);
    mc->matcher = mz_matcher;
    mc->s_ctl = s_mctl;
    
    mptn = mptn_new (control);
    mptn->mctl = mc;
    mptn->min_len = min_len;
    mptn->max_len = max_len;

    /* Need to protect mctl -- referenced from outside Scheme */
    if (s_mctl != scheme_false)
      scheme_dont_gc_ptr (s_mctl);
  }

  return mptn;
}

/* Destroy for Scheme matchers.
   Is called as (destroy mctl)
*/
static void
mz_matcher_destroy (gpointer mctl)
{
  mz_ctl_t *mc = mctl;
  mz_matcher_t *matcher = mc->matcher;
  Scheme_Object *s_mctl = mc->s_ctl;

  /* Apply the destruction procedure (if any) */
  if (matcher->destroy)
    _scheme_apply (matcher->destroy, 1, &s_mctl);

  /* Allow gc for the user data */
  if (!SCHEME_FALSEP (s_mctl))
    scheme_gc_ptr_ok (s_mctl);

  /* Free the mc structure */
  g_free (mc);
}

/* Iter start for Scheme matchers.
   Called as (iter-start mctl str vars-ectl stage data)
   Returns ictl (or #f is does not match).
*/
static gpointer
mz_matcher_iter_start (mptn_t *mptn,
		       chr *begin, chr* end,
		       mptn_var_t *vars, guint stage,
		       gpointer data)
{
  mz_ctl_t *mc = mptn->mctl;
  mz_matcher_t *matcher;
  Scheme_Object *s_mctl;
  Scheme_Object *s_str;
  Scheme_Object *s_vars;
  Scheme_Object *s_stage;
  Scheme_Object *s_data;

  Scheme_Object *s_res;

  matcher = mc->matcher;
  if (!matcher->iter_start)
    return scheme_false;

  /* Translate arguments into Scheme values */
  s_mctl = mc->s_ctl;
  s_str = scheme_make_sized_string (begin, end - begin, TRUE);
  s_vars = mz_make_vars_ectl (vars, begin);
  s_stage = scheme_make_integer (stage);
  s_data = data;

  /* Make an array of arguments, then apply the Scheme iter_start function */
  {
    Scheme_Object *s_args[5];

    s_args[0] = s_mctl;
    s_args[1] = s_str;
    s_args[2] = s_vars;
    s_args[3] = s_stage;
    s_args[4] = s_data;

    s_res = _scheme_apply (matcher->iter_start, 5, s_args);
  }

  /* Protect s_res from gc */
  if (!SCHEME_FALSEP(s_res))
    {
      mz_ctl_t *ictl;

      ictl = g_new0 (mz_ctl_t, 1);

      ictl->matcher = matcher;
      ictl->s_ctl = s_res;
      scheme_dont_gc_ptr (s_res);

      return ictl;
    }
  else
    return NULL;
}

/* Iterator step for Scheme matchers.
   Called as (iter-step ictl)
   Returns new value of ictl or #f.
*/
static gpointer
mz_matcher_iter_step (gpointer ictl)
{
  mz_ctl_t *mzictl = ictl;
  mz_matcher_t *matcher = mzictl->matcher;
  Scheme_Object *s_ictl = mzictl->s_ctl;

  Scheme_Object *s_res;

  scheme_gc_ptr_ok (s_ictl);
  
  if (!matcher->iter_step)
    {
      g_free (mzictl);
      return NULL;
    }

  s_res = _scheme_apply (matcher->iter_step, 1, &s_ictl);

  if (!SCHEME_FALSEP (s_res))
    {
      mzictl->s_ctl = s_res;
      scheme_dont_gc_ptr (s_res);
      return mzictl;
    }
  else
    {
      g_free (mzictl);
      return NULL;
    }
}

/* Ictl_free for Scheme matchers.
   Called as (ictl-free ictl).
*/
static void
mz_matcher_ictl_free (gpointer ictl)
{
  mz_ctl_t *mzictl = ictl;
  mz_matcher_t *matcher = mzictl->matcher;
  Scheme_Object *s_ictl = mzictl->s_ctl;

  scheme_gc_ptr_ok (s_ictl);

  if (matcher->ictl_free)
    _scheme_apply (matcher->ictl_free, 1, &s_ictl);

  g_free (mzictl);
}

/* Stage for Scheme matchers.
   Called as (stage ictl).
*/
static guint
mz_matcher_stage (gpointer ictl)
{
  mz_ctl_t *mzictl = ictl;
  mz_matcher_t *matcher = mzictl->matcher;
  Scheme_Object *s_ictl = mzictl->s_ctl;
  Scheme_Object *s_res;

  if (matcher->stage)
    {
      s_res = _scheme_apply (matcher->stage, 1, &s_ictl);
      if (!SCHEME_INTP (s_res))
	scheme_signal_error ("mz-matcher mptn-stage: expect integer as return value");
      
      return SCHEME_INT_VAL(s_res);
    }
  else
    return 0;
}

/* Subst for Scheme matchers.
   Called as (subst mctl ectl-vars data).
   Returns a string or #f on failure.
*/
static chr *
mz_matcher_subst (mptn_t *mptn, mptn_var_t *vars, gpointer data)
{
  mz_ctl_t *mz_mctl = mptn->mctl;
  mz_matcher_t *matcher = mz_mctl->matcher;

  Scheme_Object *s_mctl;
  Scheme_Object *s_vars;
  Scheme_Object *s_data;

  Scheme_Object *s_res;

  if (!matcher->subst)
    return NULL;

  s_mctl = mz_mctl->s_ctl;
  s_vars = mz_make_vars_ectl (vars, NULL);
  s_data = data;

  {
    Scheme_Object *s_arg[3];
    
    s_arg[0] = s_mctl;
    s_arg[1] = s_vars;
    s_arg[2] = s_data;

    s_res = _scheme_apply (matcher->subst, 3, s_arg);
  }

  if (!SCHEME_FALSEP (s_res))
    {
      if (!SCHEME_STRINGP (s_res))
	scheme_signal_error ("mz-matcher mptn-subst: expect string as return value");

      return chra_dup (SCHEME_STR_VAL(s_res));
    }
  else
    return NULL;
}

/* Print for Scheme matchers.
   Called as (print mctl port offset)
*/
static void
mz_matcher_print (mptn_t *mptn, FILE *fp, guint offset)
{
  mz_ctl_t *mz_mctl = mptn->mctl;
  mz_matcher_t *matcher = mz_mctl->matcher;
  Scheme_Object *s_mctl = mz_mctl->s_ctl;

  if (!matcher->print)
    {
      int i;
      for (i = 0; i < offset; i++)
	fputc (' ', fp);

      fputs ("mzscheme-matcher", fp);
    }
  else
    {
      Scheme_Object *s_arg[3];

      s_arg[0] = s_mctl;
      s_arg[1] = scheme_make_file_output_port (fp);
      s_arg[2] = scheme_make_integer (offset);

      _s_apply (matcher->print,3, s_arg);
    }
}

static mptn_matcher_ops_t mz_matcher = {
  mz_matcher_init,
  mz_matcher_destroy,
  mz_matcher_iter_start,
  mz_matcher_iter_step,
  mz_matcher_ictl_free,
  mz_matcher_stage,
  mz_matcher_subst,
  mz_matcher_print
};

/* Setting a matcher.
   (mptn-matcher-set! mname matcher [udata])
*/
static Scheme_Object *
mz_mptn_matcher_set_bang (int argc, Scheme_Object *argv[])
{
  mz_ctl_t *mz_ctl;
  Scheme_Object *matcher;
  Scheme_Object *udata;
  char *mname;

  g_assert (argc == 2 || argc == 3);

  if (SCHEME_SYMBOLP (argv[0]))
    mname = SCHEME_SYM_VAL (argv[0]);
  else if (SCHEME_STRINGP (argv[0]))
    mname = SCHEME_STR_VAL (argv[0]);
  else
    scheme_signal_error ("mptn-matcher-set!: expected symbol or string as the 1st argument");

  if (SCHEME_TYPE (argv[1]) == mz_matcher_type)
    matcher = argv[1];
  else
    scheme_signal_error ("mptn-matcher-set!: expected matcher as the 2nd argument");

  if (argc == 3)
    udata = argv[2];
  else
    udata = scheme_null;

  mz_ctl = scheme_malloc_eternal (sizeof (mz_ctl_t));
  mz_ctl->matcher = (mz_matcher_t*)matcher;
  mz_ctl->s_ctl = udata;

  mptn_mname_set (mz_mptn_control, mname, &mz_matcher);
  mptn_mname_param_set (mz_mptn_control, mname, mz_ctl);

  return scheme_undefined;
}


/*
  Variable value restriction.
  (mptn-var-restrict! var mptn)
   
  Var may be a symbol or a string.
*/
static Scheme_Object *
mz_mptn_var_restrict_bang (int argc, Scheme_Object *argv[])
{
  char *vname;
  mptn_t *mptn;

  g_assert (argc == 2);

  if (SCHEME_SYMBOLP (argv[0]))
    vname = SCHEME_SYM_VAL (argv[0]);
  else if (SCHEME_STRINGP (argv[0]))
    vname = SCHEME_STR_VAL (argv[0]);
  else
    scheme_signal_error ("mptn-var-restrict!: expected symbol or string as the 1st argument");

  if (SCHEME_MPTNP (argv[1]))
    mptn = SCHEME_MPTN_VAL (argv[1]);
  else
    scheme_signal_error ("mptn-var-restrict!: expected mptn as the 2nd argument");

  mptn_vname_restrict (mz_mptn_control, vname, mptn);
  return scheme_void;
}

/*
  A way to catch errors from Mptn parsing.
*/
void
mptn_parse_error (char *fmt, char *str)
{
  scheme_signal_error (fmt, str);
}


/*
  Initialization.
*/
Scheme_Object *
scheme_initialize (Scheme_Env *namespace)
{
  Scheme_Object *sym, *val;

  mz_mptn_control = mptn_control_new ();
  if (!mz_mptn_control)
    scheme_signal_error ("Could not create mptn_control");
  
  mz_mptn_type = scheme_make_type ("mptn");
  mz_ectl_type = scheme_make_type ("mptn-ectl");
  mz_matcher_type = scheme_make_type ("mptn-matcher");

  sym_valid = scheme_intern_symbol ("valid");
  sym_vars = scheme_intern_symbol ("vars");

  sym_init = scheme_intern_symbol ("init");
  sym_destroy = scheme_intern_symbol ("destroy");
  sym_iter_start = scheme_intern_symbol ("iter-start");
  sym_iter_step = scheme_intern_symbol ("iter-step");
  sym_ictl_free = scheme_intern_symbol ("ictl-free");
  sym_stage = scheme_intern_symbol ("stage");
  sym_subst = scheme_intern_symbol ("subst");
  sym_print = scheme_intern_symbol ("print");

  scheme_add_global ("mptn-parse", 
		     scheme_make_prim_w_arity (mz_mptn_parse, "mptn-parse", 1, 1),
		     namespace);
  scheme_add_global ("mptn?", 
		     scheme_make_prim_w_arity (mz_mptn_p, "mptn?", 1, 1),
		     namespace);

  scheme_add_global ("mptn-ectl?", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_p, "mptn-ectl?", 1, 1),
		     namespace);
  scheme_add_global ("mptn-ectl-valid?", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_valid_p, "mptn-ectl-valid?", 1, 1),
		     namespace);
  scheme_add_global ("mptn-ectl-ref", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_ref, "mptn-ectl-ref", 2, 2),
		     namespace);
  scheme_add_global ("mptn-ectl-has-var?", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_has_var_p, "mptn-ectl-has-var?", 2, 2),
		     namespace);
  scheme_add_global ("mptn-ectl-set!", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_set_bang, "mptn-ectl-set!", 3, 3),
		     namespace);
  scheme_add_global ("mptn-ectl->alist", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_to_alist, "mptn-ectl->alist", 1, 1),
		     namespace);
  scheme_add_global ("mptn-ectl-stage", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_stage, "mptn-ectl-stage", 1, 1),
		     namespace);
  scheme_add_global ("mptn-ectl-step!", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_step_bang, "mptn-ectl-step!", 1, 1),
		     namespace);
  scheme_add_global ("mptn-ectl-destroy!", 
		     scheme_make_prim_w_arity (mz_mptn_ectl_destroy_bang, "mptn-ectl-destroy!", 1, 1),
		     namespace);

  scheme_add_global ("mptn-exec-for-each", 
		     scheme_make_prim_w_arity (mz_mptn_exec_for_each, "mptn-exec-for-each", 4, 4),
		     namespace);
  scheme_add_global ("mptn-exec-map", 
		     scheme_make_prim_w_arity (mz_mptn_exec_map, "mptn-exec-map", 4, 4),
		     namespace);
  scheme_add_global ("mptn-exec-start", 
		     scheme_make_prim_w_arity (mz_mptn_exec_start, "mptn-exec-start", 2, 3),
		     namespace);
  scheme_add_global ("mptn-exec-start-with-ectl-and-stage", 
		     scheme_make_prim_w_arity (mz_mptn_exec_start_with_ectl_and_stage,
					       "mptn-exec-start-with-ectl-and-stage",
					       4, 5),
		     namespace);
  scheme_add_global ("mptn-subst", 
		     scheme_make_prim_w_arity (mz_mptn_subst, "mptn-subst", 2, 3),
		     namespace);

  scheme_add_global ("mptn-var-restrict!", 
		     scheme_make_prim_w_arity (mz_mptn_var_restrict_bang, "mptn-var-restrict!", 2, 2),
		     namespace);

  scheme_add_global ("mptn-matcher",
		     scheme_make_prim_w_arity (mz_mptn_make_matcher, "mptn-matcher", 0, -1),
		     namespace);
  scheme_add_global ("mptn-matcher-set!", 
		     scheme_make_prim_w_arity (mz_mptn_matcher_set_bang, "mptn-matcher-set!", 2, 3),
		     namespace);
  
  return scheme_void;
}

/*
  Reloading.
*/
Scheme_Object *
scheme_reload (Scheme_Env *namespace)
{
  /* Do nothing */
  return scheme_void;
}
